from typing import List, Union

from loguru import logger
from pika.exceptions import ChannelClosedByBroker

from petisco.base.application.chaos.check_chaos import check_chaos_publication
from petisco.base.domain.message.domain_event import DomainEvent
from petisco.base.domain.message.domain_event_bus import DomainEventBus
from petisco.extra.rabbitmq.application.message.bus.rabbitmq_message_publisher import (
    RabbitMqMessagePublisher,
)
from petisco.extra.rabbitmq.application.message.configurer.rabbitmq_message_configurer import (
    RabbitMqMessageConfigurer,
)
from petisco.extra.rabbitmq.application.message.consumer.rabbitmq_consumer_connector import (
    RabbitMqConsumerConnector,
)
from petisco.extra.rabbitmq.shared.rabbitmq_connector import RabbitMqConnector


class RabbitMqDomainEventBus(DomainEventBus):
    """
    An implementation of DomainEventBus using RabbitMQ infrastructure.
    Implementation is based on pika library.
    """

    def __init__(
        self,
        organization: str,
        service: str,
        connector: Union[
            RabbitMqConnector, RabbitMqConsumerConnector
        ] = RabbitMqConnector(),
        fallback: Union[DomainEventBus, None] = None,
    ):
        self.connector = connector
        self.exchange_name = f"{organization}.{service}"
        self.rabbitmq_key = f"publisher-{self.exchange_name}"
        self.configurer = RabbitMqMessageConfigurer(organization, service, connector)
        self.already_configured = False
        self.fallback = fallback
        self.publisher = RabbitMqMessagePublisher(self.exchange_name)

    def publish(self, domain_event: Union[DomainEvent, List[DomainEvent]]) -> None:
        """
        Publish a DomainEvent or a list of DomainEvents
        """

        meta = self.get_configured_meta()
        published_domain_event = []
        domain_events = self._check_input(domain_event)

        try:
            check_chaos_publication()
            channel = self.connector.get_channel(self.rabbitmq_key)
            for domain_event in domain_events:
                self._check_is_domain_event(domain_event)
                domain_event = domain_event.update_meta(meta)
                self.publisher.execute(channel, domain_event)
                published_domain_event.append(domain_event)
            if channel.is_open and not isinstance(
                self.connector, RabbitMqConsumerConnector
            ):
                channel.close()
        except ChannelClosedByBroker:
            unpublished_domain_events = [
                event for event in domain_events if event not in published_domain_event
            ]
            self._retry(unpublished_domain_events)
        except Exception as exc:  # noqa
            if not self.fallback:
                raise exc
            logger.opt(exception=True).error(
                f"Error publishing events ({len(domain_events)} of type {domain_events[0].get_message_type()}). Reverting to fallback. Exception:"
            )
            unpublished_domain_events = [
                event for event in domain_events if event not in published_domain_event
            ]
            self.fallback.publish(unpublished_domain_events)

    def _retry(self, domain_events: List[DomainEvent]) -> None:
        # If domain event queue is not configured, it will be configured and then try to publish again.
        if not self.already_configured:
            self.configurer.configure()
            self.already_configured = True
            self.publish(domain_events)
        elif self.fallback:
            self.fallback.publish(domain_events)

    def close(self) -> None:
        """
        Close RabbitMQ connection.
        """
        self.connector.close(self.rabbitmq_key)



import json
import re
from datetime import datetime, timezone
from typing import Any, Dict, Union, cast

from pydantic import BaseModel

from petisco.base.domain.model.uuid import Uuid
from petisco.base.domain.model.value_object import ValueObject
from petisco.base.misc.time_format import TIME_FORMAT


def get_version(config: Union[Dict[str, Any], None]) -> int:
    version = getattr(config, "version", 1) if config else 1
    return version


class MessageInfo(BaseModel):
    name: str
    version: int


class Message(BaseModel, extra="allow"):
    def model_post_init(self, __context: Any) -> None:
        if not hasattr(self, "_message_attributes"):
            attributes = dict(self)
            if "_message_type" in attributes:
                attributes.pop("_message_type")
            self._message_attributes = attributes

        if not hasattr(self, "_message_formatted_message"):
            self._message_formatted_message = None  # noqa

        if self._message_formatted_message:
            self._update_from_formatted_message()

        if not hasattr(self, "_message_id"):
            self._message_id = Uuid.v4()  # noqa

        if not hasattr(self, "_message_name"):
            self._message_name = (  # noqa
                re.sub(r"(?<!^)(?=[A-Z])", "_", self.__class__.__name__)
                .lower()
                .replace("_", ".")
            )
        if hasattr(self, "Config"):
            self._message_version = get_version(self.Config)
        else:
            self._message_version = 1  # noqa

        if not hasattr(self, "_message_occurred_on"):
            self._message_occurred_on = datetime.now(timezone.utc)  # noqa

        if not hasattr(self, "_message_attributes"):
            self._message_attributes = dict()  # noqa

        if not hasattr(self, "_message_meta"):
            self._message_meta = dict()  # noqa

        if not hasattr(self, "_message_type"):
            self._message_type = "message"  # noqa

    def add_meta(self, meta: Dict[str, Any]) -> None:
        self._message_meta = meta

    def update_meta(self, meta: Dict[str, Any]) -> "Message":
        if not meta:
            return self

        if not isinstance(meta, dict):
            raise TypeError("Message.update_meta() expect a dict")
        if hasattr(self, "_message_meta"):
            self._message_meta = {**self._message_meta, **meta}
        else:
            self._message_meta = meta
        return self

    def format(self) -> Dict[str, Dict[str, Any]]:
        data = {
            "data": {
                "id": self._message_id.value,
                "type": self._message_name,
                "type_message": self._message_type,
                "version": self._message_version,
                "occurred_on": self._message_occurred_on.strftime(TIME_FORMAT),
                "attributes": self._get_serialized_attributes(),
                "meta": self._message_meta,
            }
        }
        return data

    def format_json(self) -> str:
        return json.dumps(self.format())

    @classmethod
    def from_format(
        cls,
        formatted_message: Union[Dict[str, Any], str, bytes],
        target_type: Union[type, None] = None,
    ) -> "Message":
        if not isinstance(formatted_message, dict):
            formatted_message = json.loads(formatted_message)
        data = cast(Dict[str, Any], formatted_message.get("data"))
        attributes = (
            data.get("attributes") if data.get("attributes") is not None else dict()
        )
        target_type = target_type if target_type else cls
        message = target_type(**attributes)
        message._message_formatted_message = data
        message._update_from_formatted_message()
        return message

    def _get_serialized_attributes(self) -> Dict[str, Any]:
        attributes = {}
        for key, attribute in self._message_attributes.items():
            serialized_value = attribute
            if isinstance(attribute, (Uuid, ValueObject)):
                serialized_value = attribute.value
            if isinstance(attribute, datetime):
                serialized_value = attribute.strftime(TIME_FORMAT)
            attributes[key] = serialized_value
        return attributes

    def _update_from_formatted_message(self) -> None:
        kwargs = self._message_formatted_message
        self._message_id = Uuid(kwargs.get("id")) if kwargs.get("id") else Uuid.v4()
        self._message_name = str(kwargs.get("type"))
        self._message_version = int(kwargs.get("version", 1))
        self._message_occurred_on = (
            datetime.strptime(str(kwargs.get("occurred_on")), TIME_FORMAT).replace(
                tzinfo=timezone.utc
            )
            if kwargs.get("occurred_on")
            else datetime.now(timezone.utc)
        )

        attributes = (
            kwargs.get("attributes") if kwargs.get("attributes") is not None else dict()
        )
        self._message_attributes = cast(Dict[str, Any], attributes)
        if self._message_attributes:
            for key, value in self._message_attributes.items():
                setattr(self, key, value)

        self._message_meta = cast(Dict[str, Any], kwargs.get("meta", dict()))
        self._message_type = str(kwargs.get("type_message", self._message_type))

    def __hash__(self) -> int:
        return hash(self.format_json())

    def __eq__(self, other: Any) -> bool:
        if not isinstance(other, Message):
            return False
        return self.format() == other.format()

    def __str__(self) -> str:
        return self.__repr__()

    def __repr__(self) -> str:
        return f"{self._message_type} [{self._message_id.value} ({self._message_type}), {self._message_name} (v{self._message_version}), {self._message_occurred_on}, attributes={self._message_attributes}]"

    def get_message_id(self) -> Uuid:
        return self._message_id

    def get_message_name(self) -> str:
        return self._message_name

    def get_message_version(self) -> int:
        return self._message_version

    def get_message_occurred_on(self) -> datetime:
        return self._message_occurred_on

    def get_message_attributes(self) -> Dict[str, Any]:
        return self._message_attributes

    def get_message_meta(self) -> Dict[str, Any]:
        return self._message_meta

    def get_message_type(self) -> str:
        return self._message_type

    @classmethod
    def info(cls) -> MessageInfo:
        message_name = (  # noqa
            re.sub(r"(?<!^)(?=[A-Z])", "_", cls.__name__).lower().replace("_", ".")
        )
        message_version = 1
        if hasattr(cls, "Config") and hasattr(cls.Config, "version"):
            message_version = int(cls.Config.version)

        return MessageInfo(name=message_name, version=message_version)



from __future__ import annotations

from petisco.base.domain.message.command import Command
from petisco.base.domain.message.command_bus import CommandBus


class NotImplementedCommandBus(CommandBus):
    def dispatch(self, command: Command | list[Command]) -> None:
        commands = [command] if isinstance(command, Command) else command
        for command in commands:
            self._check_is_command(command)
            meta = self.get_configured_meta()
            _ = command.update_meta(meta)

    def close(self) -> None:
        pass



from __future__ import annotations

import re

from pydantic.main import BaseModel

from petisco.base.domain.message.command import Command
from petisco.base.domain.message.domain_event import DomainEvent
from petisco.base.domain.message.message import Message


class MessageSubscriberInfo(BaseModel):
    message_name: str
    message_version: int
    message_type: str

    @staticmethod
    def from_class_type(class_type: type[Message]) -> MessageSubscriberInfo:
        message_name = (  # noqa
            re.sub(r"(?<!^)(?=[A-Z])", "_", class_type.__name__)
            .lower()
            .replace("_", ".")
        )

        message_version = 1
        if hasattr(class_type, "Config") and hasattr(class_type.Config, "version"):
            message_version = int(class_type.Config.version)

        message_type = "message"
        if issubclass(class_type, DomainEvent):
            message_type = "domain_event"
        elif issubclass(class_type, Command):
            message_type = "command"

        return MessageSubscriberInfo(
            message_name=message_name,
            message_version=message_version,
            message_type=message_type,
        )



from typing import Any, Dict, Union

from fastapi import HTTPException
from starlette.requests import Request


class ResponseMocker:
    """
    This class mocks the router response if the client sends a valid response value through request headers
    """

    def __init__(
        self,
        header_key: str = "X-Status-Code-Mock-Response",
        responses: Union[Dict[Union[int, str], Dict[str, Any]], None] = None,
    ) -> None:
        """
        header_key: by default "X-Status-Code-Mock-Response", but you can modify with this parameter.
        responses: This will be added automatically by the FastapiApplication if you define responses to router.
        """
        self.header_key = header_key
        self.responses = responses

    def __call__(self, request: Request) -> None:
        status_code = request.headers.get(self.header_key)
        if status_code:
            status_code = int(status_code)
            detail = self.responses.get(status_code, {}) if self.responses else {}
            detail["is_mocked"] = True
            raise HTTPException(status_code=status_code, detail=detail)



import os

from loguru import logger

from petisco.extra.logger.log_message import LogMessage
from petisco.extra.logger.logger import (
    CRITICAL,
    DEBUG,
    ERROR,
    FATAL,
    INFO,
    WARN,
    WARNING,
    Logger,
)


class LoguruLogger(Logger):
    def __init__(self):
        self.logging_level = os.environ.get("PETISCO_LOGGING_LEVEL", "INFO").upper()
        logger.info(f"Set PETISCO_LOGGING_LEVEL: {self.logging_level}")

    def log(self, logging_level, log_message: LogMessage):
        message = log_message.to_dict()
        if logging_level == INFO:
            message["meta"]["level"] = "info"
            logger.info(message)
        elif logging_level == DEBUG:
            message["meta"]["level"] = "debug"
            logger.debug(message)
        elif logging_level == CRITICAL:
            message["meta"]["level"] = "critical"
            logger.critical(message)
        elif logging_level == FATAL:
            message["meta"]["level"] = "fatal"
            logger.error(message)
        elif logging_level == ERROR:
            message["meta"]["level"] = "error"
            logger.error(message)
        elif logging_level == WARN:
            message["meta"]["level"] = "warning"
            logger.warning(message)
        elif logging_level == WARNING:
            message["meta"]["level"] = "warning"
            logger.warning(message)



import time
from typing import Any

from loguru import logger
from meiga import AnyResult

from petisco.base.application.application_info import ApplicationInfo
from petisco.base.application.dependency_injection.container import Container
from petisco.base.application.middleware.middleware import Middleware
from petisco.base.application.middleware.request_responded import RequestResponded
from petisco.base.domain.message.domain_event_bus import DomainEventBus
from petisco.base.domain.value_objects.operation_type import OperationType


class RequestRespondedMiddleware(Middleware):
    """
    Middleware Implementation to generate events with information on request responses.
    """

    start_time: float

    def __init__(self) -> None:
        self.event_bus = Container.get(DomainEventBus)
        self.operation_affected = OperationType.CONTROLLER

    def get_info_id_from_input(self) -> Any:
        try:
            return self.wrapped_class_input_arguments.get("info_id", None)
        except Exception as exc:
            logger.error(
                f"Middleware error getting info_id on get_meta_from_input: {str(exc)}"
            )
        return None

    def before(self) -> None:
        self.start_time = time.time()

    def after(self, result: AnyResult) -> None:
        application_info = ApplicationInfo()

        elapsed_time = time.time() - self.start_time
        info_id = self.get_info_id_from_input()

        result_transform = result.transform()

        request_responded = RequestResponded.create(
            app_name=application_info.name,
            app_version=application_info.version,
            controller=f"{self.wrapped_class_name}",
            is_success=result.is_success,
            http_response=result_transform,
            elapsed_time=elapsed_time,
            info_id=info_id,
        )
        self.event_bus.publish(request_responded)



from __future__ import annotations

from contextlib import contextmanager
from typing import Callable, ContextManager

from elasticsearch import Elasticsearch
from loguru import logger

from petisco.base.domain.persistence.database import Database
from petisco.extra.elastic.elastic_connection import ElasticConnection


def elastic_session_scope_provider(
    session,
) -> Callable[..., ContextManager[Elasticsearch]]:
    from elasticsearch import RequestError

    @contextmanager
    def session_scope() -> ContextManager[Elasticsearch]:
        try:
            yield session
        except ConnectionRefusedError as e:
            logger.error(e)
            raise e
        except RequestError as e:
            logger.error(e)
            raise e
        except Exception as e:
            logger.error(e)
            raise e

    return session_scope


ElasticSessionScope = Callable[..., ContextManager[Elasticsearch]]


class ElasticDatabase(Database):
    session: Elasticsearch | None = None

    @staticmethod
    def local_connection_checker(alias: str | None = "test") -> ElasticDatabase:
        return ElasticDatabase(alias=alias, connection=ElasticConnection.create_local())

    def __init__(self, alias: str, connection: ElasticConnection) -> None:
        if not connection or not isinstance(connection, ElasticConnection):
            raise ConnectionError(
                "ElasticDatabase needs a valid ElasticConnection connection"
            )
        self.connection = connection
        super().__init__(alias)

    def initialize(self) -> None:
        self.session = Elasticsearch(
            self.connection.to_elastic_format(), http_auth=self.connection.http_auth
        )

    def delete(self) -> None:
        pass

    def clear_data(self) -> None:
        pass

    def is_available(self) -> bool:
        try:
            return self.session.ping()
        except Exception:  # noqa E722
            return False

    def get_session_scope(self) -> ElasticSessionScope:
        return elastic_session_scope_provider(self.session)



from contextlib import asynccontextmanager
from typing import AsyncContextManager, Callable

from loguru import logger
from sqlalchemy.exc import OperationalError
from sqlalchemy.ext.asyncio import AsyncSession


def async_sql_session_scope_provider(
    async_session_factory: Callable[[], AsyncSession]
) -> Callable[..., AsyncContextManager[AsyncSession]]:
    @asynccontextmanager
    async def session_scope() -> AsyncContextManager[AsyncSession]:
        session = async_session_factory()
        try:
            yield session
            await session.commit()
        except OperationalError as e:
            logger.error(e)
            await session.rollback()
            raise e
        except Exception as e:
            logger.error(e)
            await session.rollback()
            raise e
        finally:
            await session.close()
        # await session.delete()

    return session_scope



from abc import abstractmethod
from typing import Any

from meiga import AnyResult, NotImplementedMethodError

from petisco.base.application.use_case.meta_use_case import MetaUseCase


class UseCase(metaclass=MetaUseCase):
    """
    A base class for creating your use cases.
    """

    @abstractmethod
    def execute(self, *args: Any, **kwargs: Any) -> AnyResult:
        return NotImplementedMethodError



import json
from typing import Any, Dict, Union

from loguru import logger
from meiga import Result

from petisco.base.application.controller.http_error import HttpError
from petisco.base.domain.errors.domain_error import DomainError
from petisco.base.domain.message.domain_event import DomainEvent

REQUEST_RESPONDED_UNWRAP_ERROR = (
    {
        "error": {
            "type": "RequestRespondedUnwrapError",
            "message": "Cannot unwrap http response from decorated function",
        }
    },
)


class RequestResponded(DomainEvent):
    app_name: str
    app_version: str
    controller: str
    is_success: bool
    http_response: Dict[str, Any]
    elapsed_time: float
    info_id: Union[str, None] = None

    @staticmethod
    def create(
        app_name: str,
        app_version: str,
        controller: str,
        is_success: bool,
        http_response: Dict[str, Any],
        elapsed_time: float,
        info_id: Union[str, None] = None,
    ) -> "RequestResponded":
        app_name = app_name
        app_version = app_version
        controller = controller
        is_success = is_success
        elapsed_time = elapsed_time
        http_response = RequestResponded.set_http_response(is_success, http_response)
        info_id = info_id

        return RequestResponded(
            app_name=app_name,
            app_version=app_version,
            controller=controller,
            is_success=is_success,
            elapsed_time=elapsed_time,
            http_response=http_response,
            info_id=info_id,
        )

    @staticmethod
    def _get_content(response: Union[Dict[str, Any], str]) -> Dict[str, Any]:
        content = {"message": str(response)}

        if isinstance(response, dict):
            content = {"message": json.dumps(response)}
        elif isinstance(response, HttpError):
            content = {"message": response.detail}
        elif isinstance(response, Result):
            if isinstance(response.value, DomainError):
                content = {
                    "message": response.value.detail(),
                }
            else:
                content = {"message": str(response.value)}
        content["message"] = content["message"][:300]
        return content

    @staticmethod
    def set_http_response(is_success: bool, http_response: Any) -> Dict[str, Any]:
        try:
            _http_response = {
                "content": REQUEST_RESPONDED_UNWRAP_ERROR,
                "status_code": 500,
            }

            _http_response["content"] = RequestResponded._get_content(http_response)
            if is_success:
                _http_response["status_code"] = 200
            elif isinstance(http_response, HttpError):
                _http_response["status_code"] = http_response.status_code
            else:
                _http_response["status_code"] = 500
        except Exception as e:
            logger.error(e)
            raise e
        return _http_response



from typing import Callable

from fastapi import FastAPI

from petisco.base.application.application import Application
from petisco.extra.fastapi.application.add_controller_responses_to_response_mocker_dependencies import (
    add_controller_responses_to_response_mocker_dependencies,
)
from petisco.extra.fastapi.application.ensure_all_routers_are_async import (
    ensure_all_routers_are_async,
)


class FastApiApplication(Application):
    fastapi_configurer: Callable[[], FastAPI]
    ensure_async_routers: bool = False

    def get_app(self) -> FastAPI:
        app = self.fastapi_configurer()

        if self.ensure_async_routers is True:
            ensure_all_routers_are_async(app)

        add_controller_responses_to_response_mocker_dependencies(app)

        return app



from fastapi import FastAPI
from fastapi.routing import APIRoute
from loguru import logger


def is_async_callable(route: APIRoute) -> bool:
    endpoint = route.endpoint
    import asyncio

    if asyncio.iscoroutinefunction(endpoint):
        return True
    else:
        return False


def ensure_all_routers_are_async(app: FastAPI) -> None:
    for route in app.routes:
        if isinstance(route, APIRoute):
            if not is_async_callable(route):
                logger.error(f"Router with {route.path} is not using async definition")
                raise SystemError(
                    f"Router of {route.path} is not using async definition"
                )



from petisco.base.domain.errors.critical_error import CriticalError


class BusCannotPublish(CriticalError):
    def get_specific_detail(cls) -> str:
        return "Unable to finish all the actions of this operation"
